"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = generateHTML;
exports.getHTMLContent = getHTMLContent;

var _path = require("path");

var _ejs = _interopRequireDefault(require("ejs"));

var _mkdirp = require("mkdirp");

var _assert = _interopRequireDefault(require("assert"));

var _fs = require("fs");

var _umiPlugin = require("umi-plugin");

var _normalizeEntry = _interopRequireDefault(require("./normalizeEntry"));

var _constants = require("./constants");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const debug = require('debug')('umi-build-dev:generateHTML');

function generateHTML(opts = {}) {
  const routeConfig = opts.routeConfig,
        cwd = opts.cwd,
        config = opts.config,
        chunkToFilesMap = opts.chunkToFilesMap,
        plugins = opts.plugins,
        staticDirectory = opts.staticDirectory,
        libraryName = opts.libraryName,
        paths = opts.paths;
  const routes = Object.keys(routeConfig);
  const pagesConfig = normalizePageConfig(config.pages || {});
  routes.forEach(route => {
    const outputFilePath = (0, _path.join)(paths.absOutputPath, route.slice(1));
    const content = getHTMLContent({
      route,
      entry: routeConfig[route],
      pagesPath: paths.absPagesPath,
      root: cwd,
      pageConfig: pagesConfig && pagesConfig[route],
      chunkToFilesMap,
      plugins,
      staticDirectory,
      libraryName,
      paths,
      config
    });
    (0, _mkdirp.sync)((0, _path.dirname)(outputFilePath));
    (0, _fs.writeFileSync)(outputFilePath, content, 'utf-8');
  });
}

function normalizePageConfig(config) {
  return Object.keys(config).reduce((m, k) => {
    let newKey = k.replace(/^\.\//, '/');

    if (newKey.charAt(0) !== '/') {
      newKey = `/${newKey}`;
    }

    m[newKey] = config[k];
    return m;
  }, {});
}

function getHTMLTpl(pagesPath, rootPath, document) {
  const customTplPath = document ? (0, _path.join)(rootPath, document) : (0, _path.join)(pagesPath, 'document.ejs');
  const defaultTplPath = (0, _path.join)(__dirname, '../template/document.ejs');
  const tplPath = (0, _fs.existsSync)(customTplPath) ? customTplPath : defaultTplPath;
  return (0, _fs.readFileSync)(tplPath, 'utf-8');
}

function getFile(map, name, type) {
  if (!map) {
    return `${name}${type}`;
  }

  const files = map[name];
  (0, _assert.default)(files, `name ${name} don't exists in map ${JSON.stringify(map)}`);
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = files[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      const file = _step.value;

      if ((0, _path.extname)(file) === type) {
        return file;
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  throw new Error(`getFile failed:\nmap: ${JSON.stringify(map)}\nname: ${name}\ntype: ${type}`);
}

function getJSFiles(opts = {}) {
  const chunkToFilesMap = opts.chunkToFilesMap,
        libraryName = opts.libraryName,
        entry = opts.entry;
  const files = [getFile(chunkToFilesMap, libraryName, '.js')];

  try {
    files.push(getFile(chunkToFilesMap, `__common-${libraryName}`, '.js'));
  } catch (e) {}

  try {
    files.push(getFile(chunkToFilesMap, (0, _normalizeEntry.default)(entry), '.js'));
  } catch (e) {}

  return files;
}

function getCSSFiles(opts = {}) {
  const chunkToFilesMap = opts.chunkToFilesMap,
        libraryName = opts.libraryName,
        entry = opts.entry;
  const files = [];

  try {
    files.push(getFile(chunkToFilesMap, libraryName, '.css'));
  } catch (e) {}

  try {
    files.push(getFile(chunkToFilesMap, (0, _normalizeEntry.default)(entry), '.css'));
  } catch (e) {}

  return files;
}

function getEntry(entry) {
  if (_constants.ROUTE_FILES.indexOf((0, _path.basename)(entry)) > -1) {
    return entry.split('/').slice(0, -1).join('/');
  } else {
    return entry;
  }
}

function getHTMLContent(opts = {}) {
  const route = opts.route,
        entry = opts.entry,
        pagesPath = opts.pagesPath,
        root = opts.root,
        _opts$pageConfig = opts.pageConfig,
        pageConfig = _opts$pageConfig === void 0 ? {} : _opts$pageConfig,
        chunkToFilesMap = opts.chunkToFilesMap,
        plugins = opts.plugins,
        staticDirectory = opts.staticDirectory,
        libraryName = opts.libraryName,
        paths = opts.paths,
        config = opts.config;
  const isDev = process.env.NODE_ENV === 'development'; // 从模板生成 html

  const document = pageConfig.document,
        context = pageConfig.context;
  const tpl = getHTMLTpl(pagesPath, root, document);

  let html = _ejs.default.render(tpl, context, {
    _with: false,
    localsName: 'context'
  }); // 获取 configScript


  const resourceBaseUrl = `location.origin + window.routerBase + '${staticDirectory}/'`;
  let configScript = `
<script>
  window.routerBase = location.pathname.split('/').slice(0, -${getEntry(entry).split('/').length}).concat('').join('/');
  window.resourceBaseUrl = ${resourceBaseUrl};
</script>
`;
  configScript = (0, _umiPlugin.applyPlugins)(plugins, 'getConfigScript', configScript, {
    entry
  }); // 生成 tailBodyReplace

  let relPath = new Array(route.slice(1).split(_path.sep).length).join('../');
  relPath = relPath === '' ? './' : relPath;

  function getAssetsPath(file) {
    return `${relPath}${staticDirectory}/${file}`;
  }

  const getFilesOpts = {
    entry,
    libraryName,
    chunkToFilesMap
  };
  const cssFiles = isDev ? [] : getCSSFiles(getFilesOpts);
  const css = cssFiles.map(file => `<link rel="stylesheet" href="${getAssetsPath(file)}" />`).join('\r\n');
  const jsFiles = getJSFiles(getFilesOpts);
  const js = `
<script src="${getAssetsPath(jsFiles[0])}"></script>
${isDev ? '' : jsFiles.slice(1).map(script => `<script src="${getAssetsPath(script)}"></script>`).join('\r\n')}
  `;
  let tailBodyReplace = `
${css.trim()}
${configScript.trim()}
${js.trim()}

</body>`.trim();
  tailBodyReplace = (0, _umiPlugin.applyPlugins)(plugins, 'getTailBodyReplace', tailBodyReplace, {
    root,
    outputPath: paths.outputPath,
    staticDirectory,
    cssFiles,
    jsFiles,
    configScript
  }); // 替换 </body>

  html = html.replace('</body>', tailBodyReplace); // 插件最后处理一遍 HTML

  html = (0, _umiPlugin.applyPlugins)(plugins, 'generateHTML', html, {
    route
  });
  return `${html}\r\n`;
}